require 'socket'
require 'thread'
require 'rubygems'
require 'xmlsimple'
require 'active_record'

ActiveRecord::Base.establish_connection(
  :adapter => 'mysql',
  :host => 'localhost',
  :username => 'root',
  :password => 'saintbina',
  :database => 'accident_development',
  :socket =>  '/var/run/mysqld/mysqld.sock'
)

include Socket::Constants

class Car < ActiveRecord::Base
  has_many :accidets
end

class Accident < ActiveRecord::Base
  belongs_to :car
  has_many :accident_logs
end

class AccidentLog < ActiveRecord::Base
  belongs_to :accident
end

class AccidentImage < ActiveRecord::Base

end

class ClientHandler

	BUFFERSIZE 	= 1024
	T_XML		= 1
	T_TJPG		= 2

	def	initialize(session)
		@session = session
	end

	def change_geo_code(code)
		if /^(\d{2,3})(\d{2,2}).(\d+)[E|N]/ =~ code
		    d = $1.to_f
		    m = $2.to_f
		    s = ($3.to_f) * 0.000001 * 60

			puts (((s / 60.0) + m) / 60.0) + d

		    return (((s / 60.0) + m) / 60.0) + d
		else
    		return -1.0
		end
	end

	def change_to_date_time(date_str, gmt_str)
		if /^(\d{4,4})(\d{2,2})(\d{2,2})/ =~ date_str
			year = $1.to_i
			month = $2.to_i
			day = $3.to_i
		else
			raise Exception.new
		end
	
		if /^(\d{2,2})(\d{2,2})(\d{2,2})/ =~ gmt_str
			hour = $1.to_i
			min = $2.to_i
			sec = $3.to_i
		else
			raise Exception.new
		end

		DateTime.new(year, month, day, hour, min, sec)
	end

	def parse_accident_xml(xml_data)
	begin
		data = XmlSimple.xml_in(xml_data, {'ForceArray' => false} )
		car_number = data['HEADER']['CARNO'].to_s
		total_frame = data['HEADER']['FRAME'].to_i
  
		accident_car = Car.find_by_car_number(car_number)
  
		log_array = Array.new
  
		data['LOG'].each do |log|
			date_str =  log['DATE'].to_s
			gmt_str = log['GMT'].to_s

		    accident_time = change_to_date_time(date_str, gmt_str)
    		
			is_accident = false
			is_accident = true if log['ACCIDENT'].to_s == "1"
		
			lati = change_geo_code(log['POSITION']['LATI'].to_s)
			raise Exception.new if lati == -1
		
		    alti = change_geo_code(log['POSITION']['ALTI'].to_s)
		    raise Exception.new if alti == -1
		
			speed = log['SPEED'].to_f
		
			accel_x = log['ACCEL']['X'].to_s
			accel_y = log['ACCEL']['Y'].to_s
			accel_z = log['ACCEL']['Z'].to_s
		
			accident_log = AccidentLog.new(:lati => lati, :alti => alti, :accel_x => accel_x, 
				:accel_y => accel_y, :accel_z => accel_z, :speed => speed, :accident_time => accident_time)
		
			log_array << accident_log
		
		end
	
	  mid_accident_time = log_array[log_array.size / 2].accident_time
	  mid_lati = log_array[log_array.size / 2].lati
	  mid_alti = log_array[log_array.size / 2].alti
  
	  accident = Accident.new(:car => accident_car, :accident_time => mid_accident_time, :lati => mid_lati, :alti => mid_alti)
	  accident.save
  
	  log_array.each { |log| accident.accident_logs << log }

	  rescue Exception => e
		puts "XML ERROR"
		puts e
	  end
	end

	def parse_accident_tjpg(tjpg_data, full_size)
	begin
		data = tjpg_data
		
		total_read = 0

		car_number = data.read(18).to_s
		total_read += 18

		file_size  = data.read(4).to_s.unpack('I')[0]
		total_read += 4

		date = data.read(4).to_s.unpack('I')[0]
		total_read += 4

		accident_car = Car.find_by_car_number(car_number)

		puts "new file size = " + file_size.to_s
		puts "new date = " + date.to_s

		while total_read < full_size

			jpg_data = data.read(file_size.to_i)
			total_read += file_size.to_i
			
			puts date.to_s

			puts "total_read : " + total_read.to_s
			puts "full_size : " + full_size.to_s
			
			accident_time = Time.at(date.to_i)

			AccidentImage.create(:accident_time => accident_time, :file_size => file_size.to_i, :file_data => jpg_data, :car_id => accident_car)

			break if total_read >= full_size

			file_size_unpack = nil
			date_unpack = nil
			

			file_size_unpack = data.read(4).to_s
			total_read += 4

			date_unpack = data.read(4).to_s
			total_read += 4

			file_size = file_size_unpack.unpack('I')[0]
			date = date_unpack.unpack('I')[0]

#puts "new file size = " + file_size.to_s
#puts "new date = " + date.to_s
		end
	rescue Exeption => e
		puts e
	end
	end

	def process
		#new Connection Established
		puts "New Connection Established"
		#get auth message = ({car number} {file size})
		begin
			auth_string = @session.read(BUFFERSIZE)
		rescue Exception => e
			puts e
			return
		end

		#check auth_string is okay?
		if /^SEND\s(\S+)\s(\d+)\s(\d+)\r?\n/ =~ auth_string
			puts "Auth String is OKAY"
			car_number = $1
			file_size = $2
			transmit_mode = $3.to_i
		else
			#if auth_strig is not vaild
			#close connection and do nothing
			@session.close
			return
		end
		
		puts transmit_mode

		puts car_number
		puts file_size

		#send message to client
		#let client know server is ready to recv data
		@session.puts("ACCEPT\n")
		
		begin
			#start parse accident xml data or tjpg data
			if transmit_mode.to_i == T_XML
				rcv_data = @session.read(file_size.to_i)
				parse_accident_xml(rcv_data)
			else
				parse_accident_tjpg(@session, file_size.to_i)
			end
		rescue Exception => e
			puts e
			return
		end


		puts "COMPLETE"

		@session << "COMPLETE\n"
		@session.flush

		@sessin.close
	end
end

#main routine

server = TCPServer.new('localhost', 4321)

while ( session = server.accept )
	Thread.new(session) do |newSession|
		ClientHandler.new(newSession).process
	end
end
