/****************************************************************************
 *
 *   Copyright (c) 2006 Dave Hylands     <dhylands@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   Alternatively, this software may be distributed under the terms of BSD
 *   license.
 *
 *   See README and COPYING for more details.
 *
 ****************************************************************************/
/**
 *
 *   @file   i2c-adc.c
 *
 *   @brief  This file test ADC performance through the i2c-io interface.
 *
 ****************************************************************************/

// ---- Include Files -------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <math.h>

// ---- GPS Using Include
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "i2c-api.h"
#include "i2c-io-api.h"
#include "Log.h"

#define	FRAME		12
#define SAVETIME	(30 + 30 + 1)
#define ADCST	 	(SAVETIME * FRAME)
#define GPSFIELD	9
#define NMEABUFSZ	1024
#define XINITV		315.5
#define YINITV		360
#define ZINITV		341.5
#define ANGINITV	505
#define RANG		20
#define MAXVOLTAGE	5
#define MAXDATA		1024
#define DATPVOLT	(MAXDATA / MAXVOLTAGE)
#define VOLTPG		0.6
#define G			9.55
#define HERY		1.852
#define TRANANG		300 / 376
#define ADCNUM		5

/*struct llpoint {
  double          lat;
  double          lon;
  double          height;
  double          time;
  double          hdop;
  double          diffdelay;
  double          numsats;
  int             fix;
  int             statid;
  };*/

struct llpoint {
	double			time;
	double			lat;
	double			lon;
	double			numsats;
	double			hdop;
	double			sealv;
	double			height;
	double			speed;
	int				fix;
	int				checksum;
	char			latNW;
	char			lonSE;
	char			sealvM;
	char			valid;
};


// ---- Extern Variable

int                 i2cDev;
float				buf[ADCST][ADCNUM];
float				*sp, *ep;
char				fullFlag = 0;
double				*ssp, *sep;
int					XINITVAL = XINITV;
int					YINITVAL = YINITV;
int					ZINITVAL = ZINITV;
int					ANGINITVAL = ANGINITV;
int					RANGE = RANG;


double				sbuf[SAVETIME] = {0};
struct	llpoint		gpsBuf[SAVETIME];
struct	llpoint		*gsp, *gep;
struct 	llpoint		*lastgp;
char				fullGFlag = 0;
char 				carNo[18];

double				lastXSpeed = 0;

int                 count[4] = {0,0,0,0};
float 				x_next[4], x[4],P_next[4],P[4],K[4],Q,R,z[4];

float totalx = 0;

//int					verbose = 0;
// ---- Functions -----------------------------------------------------------
//
void				GetADC(int);
void				makeLOG();
int					checkMKLOG(int);
extern int			nmeaRead();
//***************************************************************************
/**
 *   Main entry point
 */
int main( int argc, char **argv )
{
	const char         *i2cDevName = "/dev/i2c-0";
	time_t              prevTime;
	time_t              endTime;
	int 				mux = -1;
	int					i;

	//Add Definition Alarm
	struct sigaction act;
	FILE * rfp;

	//Add Definition GDP
	int client_len;
	int client_sockfd;

	char gpsLine[NMEABUFSZ];
	char buf_get[255];

	struct sockaddr_in clientaddr;

	struct llpoint ll;

	FILE * sockin;

	//Add Definition Test

	if ((rfp = fopen("CARINFO","r")) == NULL){
		printf("Error!!!\n");
		exit(1);
	}

	fread(carNo,1,17,rfp);
	carNo[17] = '\0';

	fclose(rfp);

	LogInit( stdout );

	if ( argc >= 3 )
		RANGE = atoi(argv[2]);
	if ( argc >= 4 )
		XINITVAL = atoi(argv[3]);
	if ( argc >= 5 )
		YINITVAL = atoi(argv[4]);
	if ( argc >= 6 )
		ZINITVAL = atoi(argv[5]);
	if ( argc >= 7 )
		ANGINITVAL = atoi(argv[6]);
	if ( argc >= 2 )
		mux = atoi(argv[1]);
	else 
		mux = 0;

	if (( i2cDev = open( i2cDevName, O_RDWR )) < 0 )
	{
		LogError( "Error  opening '%s': %s\n", i2cDevName, strerror( errno ));
		exit( 1 );
	}

	//initialize kalman filter value
	for ( i = 0 ; i < 4 ; i++ ) {
		P[i] = 1;
		Q = 1.0/100000.0;
		R = 0.0001;
		x[i] = 0.0;
	}

	//initialize I2C_ADC
	I2cSetSlaveAddress( i2cDev, 0x0B, I2C_USE_CRC );

	//initialize Time
	prevTime = time( NULL );
	endTime = prevTime + mux;

	// Initialize buffer Startpoint and Endpoint
	sp = &buf[0][0];
	ep = &buf[0][0];
	ssp = &sbuf[0];
	sep = &sbuf[0];

	gsp = &gpsBuf[0];
	gep = &gpsBuf[0];
	lastgp = gep;
	printf ("%p %p \n",gep,&(gpsBuf[0]));

	// Initialize Signal
	act.sa_handler = GetADC;
	sigfillset(&(act.sa_mask));
	sigaction(SIGALRM, &act, NULL);

	//Initialize socket
	client_sockfd = socket(AF_INET, SOCK_STREAM, 0);
	clientaddr.sin_family = AF_INET;
	clientaddr.sin_addr.s_addr = inet_addr("127.0.0.1");
	clientaddr.sin_port = htons(12345);

	client_len = sizeof(clientaddr);

	if (connect(client_sockfd, (struct sockaddr *)&clientaddr, client_len) < 0)
	{
		perror("Connect error: ");
		exit(0);
	}

	// ALARM START
	ualarm(83333,83333);	

	// Initialize Verbose
	sockin = fdopen(client_sockfd,"r");

	memset (gpsLine, '\0', NMEABUFSZ);

//	int cnt  = 0;
	// Those statement need execute FRAME per seconds
	while ( time( NULL ) <= endTime || !mux)
	{
		char getType[7];
		char *token;
		char *tp;
		char sflag = 0;


		while (read(client_sockfd,buf_get,255) == -1);

		strcat(gpsLine, buf_get);

		strncpy(getType,gpsLine,6);

		if (strcmp("$GPGGA", getType) && strcmp("$GPRMC", getType)){
			tp = strstr(gpsLine, "$GPGGA");
			if (tp == NULL)
				tp = strstr(gpsLine, "$GPRMC");
			if (tp != NULL)
				strcpy(gpsLine, tp);
		}
		strncpy(getType,gpsLine,6);

		if (!strcmp ("$GPGGA", getType) && strchr(gpsLine, '\n')){
			int csum;
			char * tmpLine;
			tmpLine = (char *) malloc (sizeof(char) * (strlen(gpsLine) + 1));
			strcpy (tmpLine,gpsLine);
			// Global Positioning System Fix Data
			token = strtok(tmpLine,",\n");
			//		printf ("GPS = %s\n",token);

			// Fix UTC
			token = strtok(NULL,",\n");
			ll.time = atof(token);
			//		printf ("UTC = %s\n",token);

			// Latitude data
			token = strtok(NULL,",\n");
			ll.lat = atof(token);
			token = strtok(NULL,",\n");
			ll.latNW = *token;
			//		printf ("LAT = %lf-%c\n",ll.lat,*token);

			// Longitude data
			token = strtok(NULL,",\n");
			ll.lon = atof(token);
			token = strtok(NULL,",\n");
			ll.lonSE = *token;
			//		printf ("LAT = %lf-%c\n",ll.lon,*token);

			// Fix quality 0 = invalid, 1 = GPS fix (SPS)
			token = strtok(NULL,",\n");
			ll.valid = (char) atoi (token);
			//		printf ("Valid = %s\n",token);

			// Number of satellites being tracked
			token = strtok(NULL,",\n");
			ll.numsats = atof(token);
			//		printf ("NUMSAT = %s\n",token);

			// Horizontal dilution of position
			token = strtok(NULL,",\n");
			ll.hdop = atof(token);
			//		printf ("HDOP= %s\n",token);

			// Altitude, Meters, above mean sea level
			token = strtok(NULL,",\n");
			ll.sealv = atof(token);
			token = strtok(NULL,",\n");
			ll.sealvM = *token;
			//		printf ("SEA Lv = %lf-%c\n",ll.sealv, *token);

			// the checksum data, always begins with *
			token = strtok(NULL,"*\n");
			token = strtok(NULL,"\n");

			ll.checksum = strtol(token, NULL, 16);

			csum = 0;
			for (token = gpsLine + 1; *token; token++){
				if (*token == '*') // delimiter
					break;
				csum ^= *token;
			}
			if (csum != ll.checksum) {
				//Checksum Error
				printf ("CheckSum Error!\n");
				//MAKE ERROR EXEC
			}
			//printf ("Checksum = %d\n", ll.checksum);

			//Speed Parsing
			if (ll.valid == 1)
				printf("GetLine = %s\n",gpsLine);
			strcpy(tmpLine, strchr(gpsLine,'\n') + 1);
			memset(gpsLine, '\0', NMEABUFSZ);
			strcpy(gpsLine, tmpLine);
			free(tmpLine);
			//			printf("Remain String [%s]\n",gpsLine);

		} else if(!strcmp ("$GPRMC", getType) && strchr(gpsLine, '\n')){
			char * tmpLine;
			tmpLine = (char *) malloc (sizeof(char) * (strlen(gpsLine) + 1));
			strcpy (tmpLine,gpsLine);
			// Global Positioning System Fix Data
			token = strtok(tmpLine,",\n");
			//		printf ("GPS = %s\n",token);
			token = strtok(NULL,",\n");
			token = strtok(NULL,",\n");
			token = strtok(NULL,",\n");
			token = strtok(NULL,",\n");
			token = strtok(NULL,",\n");
			token = strtok(NULL,",\n");
			token = strtok(NULL,",\n");
			ll.speed = atof(token);

			if (ll.valid == 1){
				printf("GPS Speed = %lf\n",ll.speed);
				printf("GetLine = %s\n",gpsLine);
			}

			strcpy(tmpLine, strchr(gpsLine,'\n') + 1);
			memset(gpsLine, '\0', NMEABUFSZ);
			strcpy(gpsLine, tmpLine);
			free(tmpLine);
			sflag = 1;

		}
		else 
			continue;

		if (ll.valid){
			if (!sflag){
				gep -> lat = ll.lat;
				gep -> lon = ll.lon;
				gep -> height = ll.height;
				gep -> time = ll.time;
				gep -> hdop = ll.hdop;
				gep -> numsats = ll.numsats;
				gep -> fix = ll.fix;
				gep -> sealv = ll.sealv;
				gep -> latNW = ll.latNW;
				gep -> lonSE = ll.lonSE;
				gep -> sealvM = ll.sealvM;
				gep -> valid = ll.valid;
			}else{
				gep -> speed = ll.speed;


				lastgp = gep;

				gep++;

				lastXSpeed = ll.speed * HERY;
				printf("GPS Speed Sync!!\n");
				*sep = lastXSpeed;

//				gep -> lat = cnt++;
				sep++;

				if (gep == &(gpsBuf[SAVETIME])){
					gep = &(gpsBuf[0]);
					sep = &(sbuf[0]);
					fullGFlag = 1;
				}

				if ( fullGFlag ){
					gsp ++;
					ssp ++;
				}

				if (gsp == &(gpsBuf[SAVETIME])){
					gsp = &(gpsBuf[0]);
					ssp = &(sbuf[0]);
				}
			}
		}else{
			if (!sflag){
				gep -> lat = lastgp -> lat;
				gep -> lon = lastgp -> lon;
				gep -> height = lastgp -> height;
				gep -> time = lastgp -> time + 1;
				gep -> hdop = lastgp -> hdop;
				gep -> numsats = lastgp -> numsats;
				gep -> fix = lastgp -> fix;
				gep -> sealv = lastgp -> sealv;
				gep -> latNW = 'N';
				gep -> lonSE = 'E';
				gep -> sealvM = lastgp -> sealvM;
				gep -> valid = 0;
			}else{
				gep -> speed = lastgp -> speed;

//				gep -> lat = cnt++;
				lastgp = gep;

				gep++;
				*sep = lastXSpeed;

				sep++;

				if (gep == &(gpsBuf[SAVETIME])){
					gep = &(gpsBuf[0]);
					sep = &(sbuf[0]);
					fullGFlag = 1;
				}

				if ( fullGFlag ){
					gsp ++;
					ssp ++;
				}

				if (gsp == &(gpsBuf[SAVETIME])){
					gsp = &(gpsBuf[0]);
					ssp = &(sbuf[0]);
				}
			}

//			printf("%lf\n",gep->lat);
		}

		if ( checkMKLOG(sflag) ){
			makeLOG();
			sflag = 0;
		}
	}
	ualarm(0,0);

	close(client_sockfd);

	// END

	close( i2cDev );

	return 0;
}


void GetADC(int arg){
	uint16_t    adcVal;
	int 		i;
	//	static	double cnt = 0;

	for ( i = 0 ; i < 4 ; i++ ) {
		if ( I2C_IO_GetADC( i2cDev, i, &adcVal ))
		{
			float temp;
			x_next[i] = x[i];
			P_next[i] = P[i] + Q;
			K[i] = P_next[i]/(P_next[i]+R);
			z[i] = -0.37727+adcVal;
			x[i] = x_next[i] + K[i] * (z[i] - x_next[i]);
			P[i] = (1-K[i]) * P_next[i];

			switch (i){
				case 0:
					temp = XINITVAL;
					break;
				case 1:
					temp = YINITVAL;
					break;
				case 2:
					temp = ZINITVAL;
					break;
				default:
					*(ep + i) = (x[i] - ANGINITVAL) * TRANANG;
//					printf("ANG = %lf\n",*(ep + i));
					continue;
			}

			*(ep + i) = (x[i] - temp)/(1024/5)/VOLTPG * G;

			if (i == 2){
	//			printf("X = %lfm/s2, Y = %lfm/s2, Z = %lfm/s2 ",*(ep),*(ep + 1),*(ep + 2));
				printf("X = %lfm/s2\n",*(ep));
				*(sep) = lastXSpeed + ((*ep)/FRAME) * 6 * 6 / 10;
				lastXSpeed = *(sep);
//				printf ("SPEED = %f km/h \n",lastXSpeed);
			}
			count[i]++;
		}
		else
		{
			LogError( "Failed to retrieve ADC value\n" );
			break;
		}
	}
	ep += ADCNUM;
	//	cnt++;

	if (ep == &(buf[ADCST][0])){
		ep = &buf[0][0];
		fullFlag = 1;
	}

	if ( fullFlag ){
		sp += ADCNUM;
	}

	if (sp == &(buf[ADCST][0])){
		sp = &buf[0][0];
	}

}
void makeLOG (){
	FILE * ifp;
	char fileName[20];
	int i,j;
	time_t timer;
	struct tm *t;
	float * ttsp;
	int jtmp = 0;
	char acciflag = 0;

	float *tsp = ep + ADCNUM;
	double *tssp = sep + 1;
	struct llpoint *tgsp = gep + 1;

	if (!fullFlag)
		tsp = sp;

	if (!fullGFlag){
		tssp = ssp; 
		tgsp = gsp;
	}

	if (tsp > &(buf[ADCST - 1][0]))
		tsp = &(buf[0][0]);
	if (tssp > &(sbuf[SAVETIME - 1]))
		tssp = &(sbuf[0]);
	if (tgsp > &(gpsBuf[SAVETIME - 1]))
		tgsp = &(gpsBuf[0]);

	system("rdate -s time.bora.net 2> /dev/null");
	sprintf (fileName,"./log/log%ld.xml",time(NULL));
	if ((ifp = fopen(fileName, "w")) != NULL){
		fprintf(ifp,"<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
		fprintf(ifp,"<!DOCTYPE\tBLACKBOX\n");
		fprintf(ifp,"[<!ELEMENT BLACKBOX\t(HEADER,LOG*)>\n");
		fprintf(ifp,"<!ELEMENT HEADER\t(FRAME,CARNO)>\n");
		fprintf(ifp,"<!ELEMENT FRAME\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT CARNO\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT LOG\t(DATE, GMT, ACCEL, ACCIDENT, SPEED, ANGLE_A, POSITION)>\n");
		fprintf(ifp,"<!ELEMENT DATE\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT GMT\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT ACCEL\t(X,Y,Z)>\n");
		fprintf(ifp,"<!ELEMENT X\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT Y\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT Z\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT ACCIDENT\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT SPEED\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT ANGLE_A\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT POSITION\t(VAL, LATI, ALTI)>\n");
		fprintf(ifp,"<!ELEMENT VAL\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT LATI\t(#PCDATA)>\n");
		fprintf(ifp,"<!ELEMENT ALTI\t(#PCDATA)>\n");
		fprintf(ifp,"]>\n\n");

		fprintf(ifp,"<BLACKBOX>\n");
		fprintf(ifp,"\t<HEADER>\n");
		fprintf(ifp,"\t\t<FRAME>%d</FRAME>\n",FRAME);
		fprintf(ifp,"\t\t<CARNO>%s</CARNO>\n",carNo);
		fprintf(ifp,"\t</HEADER>\n");

		for (i = 0; i < SAVETIME - 1; i++){
			timer = time(NULL);
			t = localtime(&timer);
			timer -= (SAVETIME - i - 1);
			fprintf(ifp,"\n\t<LOG>\n");
//			fprintf(ifp,"\t\t<DATE>%04d%02d%02d</DATE>\n",t->tm_year + 1900,t->tm_mon + 1,t->tm_mday);
			fprintf(ifp,"\t\t<DATE>20071004</DATE>\n");
			fprintf(ifp,"\t\t<GMT>%06.0lf</GMT>\n",tgsp -> time);
			fprintf(ifp,"\t\t<ACCEL>\n");

			fprintf(ifp,"\t\t\t<X>");
			jtmp = 0;
			for (j = 0; j < FRAME - 1; j++){
				if (tsp + (j * ADCNUM) >= &(buf[ADCST][0])){
					ttsp = &(buf[0][0]) + ADCNUM * (j - 1 - jtmp);
				}
				else{
					ttsp = tsp + (j * ADCNUM);
					jtmp = j;
				}
				fprintf(ifp,"%f,",*(ttsp));
			}
			if (tsp + (j * ADCNUM) >= &(buf[ADCST][0])){
				ttsp = &(buf[0][0]) + ADCNUM * (j - 1 - jtmp);
			}else
				ttsp = tsp + (j * ADCNUM);
			fprintf(ifp,"%f</X>\n",*(ttsp));

			fprintf(ifp,"\t\t\t<Y>");
			jtmp = 0;
			for (j = 0; j < FRAME - 1; j++){
				if (tsp + (j * ADCNUM) >= &(buf[ADCST][0])){
					ttsp = &(buf[0][0]) + ADCNUM * (j - 1 - jtmp);
				}
				else{
					ttsp = tsp + (j * ADCNUM);
					jtmp = j;
				}
				fprintf(ifp,"%f,",*(ttsp + 1));
			}
			if (tsp + (j * ADCNUM) >= &(buf[ADCST][0])){
				ttsp = &(buf[0][0]) + ADCNUM * (j - 1 - jtmp);
			}else
				ttsp = tsp + (j * ADCNUM);
			fprintf(ifp,"%f</Y>\n",*(ttsp + 1));

			fprintf(ifp,"\t\t\t<Z>");
			jtmp = 0;
			for (j = 0; j < FRAME - 1; j++){
				if (tsp + (j * ADCNUM) >= &(buf[ADCST][0])){
					ttsp = &(buf[0][0]) + ADCNUM * (j - 1 - jtmp);
				}
				else{
					ttsp = tsp + (j * ADCNUM);
					jtmp = j;
				}
				fprintf(ifp,"%f,",*(ttsp + 2));
			}
			if (tsp + (j * ADCNUM) >= &(buf[ADCST][0])){
				ttsp = &(buf[0][0]) + ADCNUM * (j - 1 - jtmp);
			}else
				ttsp = tsp + (j * ADCNUM);
			fprintf(ifp,"%f</Z>\n",*(ttsp + 2));

			fprintf(ifp,"\t\t</ACCEL>\n");

			fprintf(ifp,"\t\t<ACCIDENT>");
			jtmp = 0;
			acciflag = 0;
			for (j = 0; j < FRAME; j++){
				if (tsp + (j * ADCNUM) >= &(buf[ADCST][0])){
					ttsp = &(buf[0][0]) + ADCNUM * (j - 1 - jtmp);
				}
				else{
					ttsp = tsp + (j * ADCNUM);
					jtmp = j;
				}
				if (*(ttsp + 4) > 0) acciflag = 1;
			}
			fprintf(ifp,"%d</ACCIDENT>\n",acciflag);


			fprintf(ifp,"\t\t<SPEED>%lf</SPEED>\n",*tssp);
			fprintf(ifp,"\t\t<ANGLE_A>");

			jtmp = 0;
			for (j = 0; j < FRAME - 1; j++){
				if (tsp + (j * ADCNUM) >= &(buf[ADCST][0])){
					ttsp = &(buf[0][0]) + ADCNUM * (j - 1 - jtmp);
				}
				else{
					ttsp = tsp + (j * ADCNUM);
					jtmp = j;
				}
				fprintf(ifp,"%f,",*(ttsp + 3));
			}
			if (tsp + (j * ADCNUM) >= &(buf[ADCST][0])){
				ttsp = &(buf[0][0]) + ADCNUM * (j - 1 - jtmp);
			}else
				ttsp = tsp + (j * ADCNUM);

			fprintf(ifp,"%f</ANGLE_A>\n",*(ttsp + 3));

			fprintf(ifp,"\t\t<POSITION>\n");
			fprintf(ifp,"\t\t\t<VAL>%d</VAL>\n",tgsp -> valid);
			fprintf(ifp,"\t\t\t<LATI>%lf%c</LATI>\n",tgsp -> lat, tgsp -> latNW);
			fprintf(ifp,"\t\t\t<ALTI>%lf%c</ALTI>\n",tgsp -> lon, tgsp -> lonSE);
			fprintf(ifp,"\t\t</POSITION>\n");
			fprintf(ifp,"\t</LOG>\n");

			tsp += (ADCNUM * FRAME);

			if (tsp >= &(buf[ADCST][0])){
				tsp = &buf[0][0];
			}

			tssp ++;
			tgsp ++;

			if (tgsp == &(gpsBuf[SAVETIME])){
				tgsp = &(gpsBuf[0]);
				tssp = &(sbuf[0]);
			}
		}	
		fprintf(ifp,"</BLACKBOX>\n");
		fclose(ifp);
	}
}

int checkMKLOG(int sflag){
	static int cnt = -1;
	static int flag = 0;

	int i;
	float *lastestVal = (ep - (ADCNUM * FRAME));

	if (!sflag)
		return 0;

	for (i = 0; i < FRAME && !flag; i++){
		if ( abs(*lastestVal) >  ((RANGE) * G/(1024/5)/VOLTPG) && *lastestVal < 0) { // Breaking
			printf ("%lf  ",*lastestVal);
			*(lastestVal + 4) = 1;
			flag = 1;
			break;
		} else 
			*(lastestVal + 4) = 0;

		// Rounding Skip
		lastestVal+= ADCNUM;
	}

	if (flag){
		cnt++;
		if (cnt == 30){
			cnt = -1;
			flag = 0;
			return 1;
		}
		printf("Waiting Time Remain %d \n",30 - cnt);
	}

	return 0;
}
