#include "logger.h"
#include "i2c-api.h"
#include "i2c-io-api.h"
#include "Log.h"

using namespace std;

// ---- About GPS Variable ---- //
struct llpoint gps_buf;

// ---- About Sensor (i2c) Variable ---- //
int 				i2cDev;
// for kalman filter
int                 count[NSENSOR] = {0,};
float               x_next[NSENSOR], x[NSENSOR],P_next[NSENSOR],P[NSENSOR],K[NSENSOR],Q,R,z[NSENSOR];
// Last Speed
double				lastXSpeed;
double				XINITVAL = DEFXINIT;
double				YINITVAL = DEFYINIT;
double				ZINITVAL = DEFZINIT;
double				RANGE = DEFRANGE;

// ---- About Video Stream Variable ---- //
// signal fresh frames 
pthread_mutex_t		db        = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t		db_update = PTHREAD_COND_INITIALIZER;
pthread_mutex_t		fps       = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t		fps_update = PTHREAD_COND_INITIALIZER;
pthread_mutex_t		gps 	  = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t		gps_update = PTHREAD_COND_INITIALIZER;
// videobuffer struct
struct				vdIn *videoBuffer;
// videoBuffer
unsigned char		*g_buf = NULL;
// videosize
size_t				g_size;

// ---- About make_log Variable ---- //
// Queue
deque<ONESEC>		DQ;
deque<CAMFPS>		CDQ;
// Ready node and index
ONESEC				node;
int					nowNodeIdx;
// Make log Flag
bool				isLoging = false;
// First Call Flag
bool				firstTime = true;
// Accident patten flag
bool				acciFlag = false;
int					acciWaitingTime = SAVETIME/2;
// Develop Log Level
int 				sensorLogLevel = 0;
int 				camLogLevel = 0;
// Car identification number
string 				car_number;
string				xml_log_path;
string				fps_log_path;
// now Time
int now_time;
// Process Running Flag
bool stop;

/* the single writer thread */
void *cam_thread( void *arg ) {
	while( !stop ) {
		/* grab a frame */
		while( isLoging ) sleep(1);
		if( uvcGrab(videoBuffer) < 0 ) {
			cout << "Error grabbing" << endl;
		//	exit(1);
		}

		// Log Print
		if (camLogLevel > 1)
			cout << "Capture!" << endl;

		/* copy frame to global buffer */
		pthread_mutex_lock( &db );

		g_size = videoBuffer->buf.bytesused;
		memcpy(g_buf, videoBuffer->tmpbuffer, videoBuffer->buf.bytesused);

		/* signal fresh_frame */
		pthread_cond_broadcast(&db_update);
		pthread_mutex_unlock( &db );

		/* only use usleep if the fps is below 5, otherwise the overhead is too long */
		if ( videoBuffer->fps < 5 ) {
			usleep(1000*1000/videoBuffer->fps);
		}
	}
	return NULL;
}

// Process per one second
void *sec_thread(void * arg){
	struct ONESEC trashNode;
	size_t frame_size;

	while (!stop){
		if (!firstTime){
			if (DQ.size() >= SAVETIME){
				trashNode = DQ.front();
				free(trashNode.video);
				DQ.pop_front();
			}
			DQ.push_back(node);
		}else{
			firstTime = false;
		}

		/* Allocate video buffer for second */
		node.video = (unsigned char *) calloc(1, (size_t)videoBuffer->framesizeIn);
		node.accident_flag = 0;
		memset(&node.gps_data, 0, sizeof(llpoint));
		node.gps_data.latNW = '0';
		node.gps_data.lonSE = '0';
		node.gps_data.sealvM = '0';
		node.gps_data.valid = '0';

		node.speed = lastXSpeed;
		node.time = now_time++;
		// Get GPS DATA

		/* wait for fresh frames */
		pthread_cond_wait(&gps_update, &gps);
		node.gps_data = gps_buf;
		pthread_mutex_unlock( &gps );
		
		// Log Print
		if (camLogLevel + sensorLogLevel > 0)
			cout << " " << "GPS DATA : " << " " << node.gps_data.time << " " << node.gps_data.lat << " " << 
				node.gps_data.latNW << " " << node.gps_data.lon << " " << node.gps_data.lonSE << " " << 
				node.gps_data.fix << " " << node.gps_data.numsats << " " << node.gps_data.hdop <<
				node.gps_data.sealv << " " << node.gps_data.sealvM << " " << node.gps_data.height << " " << 
				node.gps_data.speed << " " << endl;

		// GPS correction!
		if (node.gps_data.valid != '1'){
			node.speed = node.gps_data.speed;
			node.time = (int)node.gps_data.time;
			now_time = node.time;
		}

		/* wait for fresh frames */
		pthread_cond_wait(&db_update, &db);

		/* read buffer */
		frame_size = g_size;
		memcpy(node.video, g_buf, frame_size);
		node.video_length = frame_size;

		pthread_mutex_unlock( &db );

		// if accident occur
		if (acciFlag){
			acciWaitingTime--;
			// Log Print
			if (sensorLogLevel + camLogLevel > 0)
				cout << acciWaitingTime << " seconds left" << endl;
		}
		/* Complete wating remain seconds */
		if (acciWaitingTime == 0){
			// make_log start!
			make_log();

			// Accident Flag initialize
			acciWaitingTime = SAVETIME/2;
			acciFlag = false;
			if (sensorLogLevel + camLogLevel > 0)
				cout << "Make Log Complete!" << endl;
		}
	}

	return NULL;
}

void make_log(){
	string log_dir = "";
	ostringstream outs;

	isLoging = true;

	// Date Sync..
	if (camLogLevel + sensorLogLevel > 0)
		cout << "Date Sync..." << endl;

	// Make log directory
	outs << LOGPATH << xml_log_path << now_time;
	log_dir = outs.str() + ".";
	// Make XML log
	make_XML_log(log_dir);
	// Make Image log 
	make_capture_log(log_dir);

	outs.str("");
	outs << LOGPATH << fps_log_path << now_time;
	// Make FPS Log Directory
	log_dir = outs.str() + "/";
	if (camLogLevel + sensorLogLevel > 0)
		cout << "Make Dirctory.. ['" << log_dir << "']" << endl;
	mkdir(log_dir.c_str(), S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
	// Make FPS log
	make_fps_log(log_dir);

	isLoging = false;
}

// 1/fps second
void div_sec(int arg){
	static int cnt = 0;
	static int camcnt = 0;

	if (isLoging) // If it`s making log
		return;

	nowNodeIdx = cnt;
	get_adc();

	cnt++;
	camcnt++;

	if (cnt >= FPS){ 
		cnt = 0;
	}

	if (camcnt >= FPS / VIDEOFPS){
		camcnt = 0;
		// CAM FPS Processing
		pthread_cond_broadcast(&fps_update);
	}
}

void *fps_thread(void *){
	CAMFPS fps_node, trashNode;
	size_t frame_size;

	while (!stop){
		/* Allocate video buffer for second */
		fps_node.video = (unsigned char *)calloc(1, (size_t)videoBuffer->framesizeIn);
		/* wait for fresh frames */
		pthread_cond_wait(&db_update, &db);
		/* read buffer */
		frame_size = g_size;
		memcpy(fps_node.video, g_buf, frame_size);
		fps_node.video_length = frame_size;

		if (camLogLevel > 2)
			cout << "FPS Cam Capture" << endl;

		pthread_mutex_unlock( &db );

		if (CDQ.size() >= (VIDEOFPS) * SAVETIME){
			trashNode = CDQ.front();
			free(trashNode.video);
			CDQ.pop_front();
		}
		CDQ.push_back(fps_node);
		
		pthread_cond_wait(&fps_update, &fps);
	}
	return NULL;
}

// Get data from AC
void get_adc(){
	uint16_t adcVal;
	for (int i = 0 ; i < NSENSOR ; i++ ) {
		if ( I2C_IO_GetADC( i2cDev, i, &adcVal )) {
			double temp = 0;
			x_next[i] = x[i];
			P_next[i] = P[i] + Q;
			K[i] = P_next[i]/(P_next[i]+R);
			z[i] = -0.37727+adcVal;
			x[i] = x_next[i] + K[i] * (z[i] - x_next[i]);
			P[i] = (1-K[i]) * P_next[i];

			switch(i){
				case 0:
					temp = XINITVAL;
					break;
				case 1:
					temp = YINITVAL;
					break;
				case 2:
					temp = ZINITVAL;
					break;
				default:	// maybe use can use angle accelerator in here
					break; 
			} 
			// Calcurate Acceleration
			node.frame[nowNodeIdx][i] = (x[i] - temp) * MPSFOMULA; 
		} 
	} 
	if (sensorLogLevel > 0){ // Log Print! 
		cout << "X = " << node.frame[nowNodeIdx][0] << " Y = " << node.frame[nowNodeIdx][1] << " Z = " << node.frame[nowNodeIdx][2] << endl; 
	} 
	// Speed Correction by using sensor value 
	lastXSpeed += (node.frame[nowNodeIdx][0]/FPS) * 6 * 6 / 10;

	// Acccident ?? 
	if (is_accident_patten(node.frame[nowNodeIdx])){
		node.accident_flag = 1;
		acciFlag = true;
	}
}

// Make capture Log
void make_capture_log(string log_dir){
	int fd;
	int log_cnt = 1;
	string fileName = log_dir;

	//fileName Setting
	fileName += CAMLOGEXT;

	if (camLogLevel > 0){
		cout << "Capture log file name : " << fileName << endl;
	}
	// Capture log file open
	fd = open(fileName.c_str(), O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);

	// write car information
	write(fd, car_number.c_str(), sizeof(char) * CARNOLENGTH);
	for (deque<ONESEC> :: iterator it = DQ.begin() ; it != DQ.end() ; it++){
		// Log Print
		if (camLogLevel > 2){
			cout << log_cnt << "th Image Append!" << endl;
			log_cnt++;
		}
		// write video Length
		size_t real_size = picture_size(it -> video, it -> video_length);
		write(fd, &real_size, sizeof(size_t));
		write(fd, &it -> time, sizeof(time_t));
		print_picture(fd, it -> video, it -> video_length);
		cout << it -> video_length << endl;
	}
	// Capture log file close
	close(fd);
}

void make_fps_log(string log_dir){
	string fileName = log_dir;
	ostringstream outs;
	int cnt = 0;
	int fd;

	cnt = 1;
	for (deque<CAMFPS> :: iterator it = CDQ.begin() ; it != CDQ.end() ; it++){
		// Log Print
		if (camLogLevel > 2){
			cout << cnt << "th Image Append!" << endl;
		}

		outs.str("");
		outs << cnt << "." << FPSLOGEXT;
		fileName = log_dir + outs.str();
		cout << fileName << endl;
		fd = open(fileName.c_str(), O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
		// make Log
		print_picture(fd, it -> video, it -> video_length);
		close(fd);

		cnt++;
	}
}

void make_XML_log(string log_dir){
	int log_cnt = 1;
	struct tm *sensor_time;
	string fileName = log_dir;

	//fileName Setting
	fileName += XMLLOGEXT;

	if (sensorLogLevel > 0){
		cout << "XML log file name : " << fileName << endl;
	}

	ofstream fout(fileName.c_str());
	// XML Header
	fout << "<?xml version=\"1.0\" encoding=\"utf-8\"?>" << endl;
	fout << "<!DOCTYPE\tBLACKBOX" << endl;
	fout << "[<!ELEMENT BLACKBOX\t(HEADER,LOG*)>" << endl;
	fout << "<!ELEMENT HEADER\t(FRAME,CARNO)>" << endl;
	fout << "<!ELEMENT FRAME\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT CARNO\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT LOG\t(DATE, GMT, ACCEL, ACCIDENT, SPEED, ANGLE_A, POSITION)>" << endl;
	fout << "<!ELEMENT DATE\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT GMT\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT ACCEL\t(X,Y,Z)>" << endl;
	fout << "<!ELEMENT X\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT Y\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT Z\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT ACCIDENT\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT SPEED\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT ANGLE_A\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT POSITION\t(VAL, LATI, ALTI)>" << endl;
	fout << "<!ELEMENT VAL\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT LATI\t(#PCDATA)>" << endl;
	fout << "<!ELEMENT ALTI\t(#PCDATA)>" << endl;
	fout << "]>" << endl << endl;

	// Car information Log
	fout << "<BLACKBOX>" << endl;
	fout << "\t<HEADER>" << endl;
	fout << "\t\t<FRAME>" << FPS << "</FRAME>" << endl;
	fout << "\t\t<CARNO>" << car_number << "</CARNO>" << endl;
	fout << "\t</HEADER>" << endl;

	// 1 Second Data. (GPS + Sensoring)
	for (deque<ONESEC> :: iterator it = DQ.begin() ; it != DQ.end() ; it++){
		// Log Print
		if (sensorLogLevel > 2){
			cout << log_cnt << "th Log Append!" << endl;
			log_cnt++;
		}

		fout << "\t<LOG>" << endl;
		{
			// Date Log
			sensor_time = localtime(&it -> time);
			// To-Do : Complete Format yyyymmdd - i don`t want to use sprintf
			fout << "\t\t<DATE>";
			fout << sensor_time -> tm_year + 1900;
			if (sensor_time -> tm_mon + 1 < 10)
				fout << 0;
			fout << sensor_time -> tm_mon + 1;
			if (sensor_time -> tm_mday < 10)
				fout << 0;
			fout << sensor_time -> tm_mday;
			fout << "</DATE>" << endl;

			// GMT Log
			fout << "\t\t<GMT>";
			fout << it -> time;
			fout << "</GMT>" << endl;

			// Sensor Log
			fout << "\t\t<ACCEL>" << endl;
			{
				// X axis
				fout << "\t\t\t<X>";
				fout << it -> frame[0][0];
				for (int i = 1 ; i < FPS ; i++){
					fout << "," << it -> frame[0][i];
				}
				fout << "</X>" << endl;

				// Y axis
				fout << "\t\t\t<Y>";
				fout << it -> frame[1][0];
				for (int i = 1 ; i < FPS ; i++){
					fout << "," << it -> frame[1][i];
				}
				fout << "</Y>" << endl;

				// Z axis
				fout << "\t\t\t<Z>";
				fout << it -> frame[2][0];
				for (int i = 1 ; i < FPS ; i++){
					fout << "," << it -> frame[2][i];
				}
				fout << "</Z>" << endl;
			}
			fout << "\t\t</ACCEL>" << endl;

			// ACCIDENT Flag this is representing accident timing
			fout << "\t\t<ACCIDENT>" << it -> accident_flag << "</ACCIDENT>" << endl;

			// Speed
			fout << "\t\t<SPEED>" << it -> speed << "</SPEED>" << endl;

			// Angle data Skip!
			// Position data from GPS
			fout << "\t\t<POSITION>" << endl;
			{
				fout << "\t\t\t<VAL>" << it -> gps_data.valid << "</VAL>" << endl;
				fout << "\t\t\t<LATI>" << it -> gps_data.lat << it -> gps_data.latNW << "</LATI>" << endl;
				fout << "\t\t\t<ALTI>" << it -> gps_data.lon << it -> gps_data.lonSE << "</ALTI>" << endl;
			}
			fout << "\t\t</POSITION>" << endl;
		}
		fout << "\t</LOG>" << endl;
	}

	// Footer
	fout << "</BLACKBOX>" << endl;

	fout.close();
}

bool is_accident_patten(double accels[3]){
	// X axis
	// Over Range Break?	
	if ( accels[0] > ((RANGE) * MPSFOMULA)) {
		if (sensorLogLevel > 0)
			cout << "Accident Patten! X axis Acceleration = " << accels[0] << endl;
		return true;
	}

	// Y axis
	// Over Range?
	if ( fabs(accels[1]) > ((RANGE) * MPSFOMULA) ){
		if (sensorLogLevel > 0)
			cout << "Accident Patten! Y axis Acceleration = " << accels[1] << endl;
		return true;
	}

	// Z axis
	// Free Fall
	//	if ( accels[2] <= 0 ){
	//		if (sensorLogLevel > 0)
	//			cout << "Accident Patten! Z axis Acceleration = " << accels[2] << endl;
	//		return true;
	//	}

	return false;
}

int main(int argc, char *argv[]){
	pthread_t cam_t, sec_t, fps_t, gps_t;
	struct sigaction act;
	const char *i2cDevName = I2CDEV;

	// Development mode setting
	for (int i = 1 ; i < argc - 1 ; i++){
		if (!strcmp(argv[i], "-s")){
			sensorLogLevel = atoi(argv[++i]);
			cout << "Sensor Log Mode - Level " << sensorLogLevel << endl;
		}else if (!strcmp(argv[i], "-c")){
			camLogLevel = atoi(argv[++i]);
			cout << "Cam Log Mode - Level " << camLogLevel << endl;
		}else if (!strcmp(argv[i], "-x")){
			XINITVAL = atoi(argv[++i]);
			cout << "X Init Value Resetup to " << XINITVAL << endl;
		}else if (!strcmp(argv[i], "-y")){
			YINITVAL = atoi(argv[++i]);
			cout << "Y Init Value Resetup to " << YINITVAL << endl;
		}else if (!strcmp(argv[i], "-z")){
			ZINITVAL = atoi(argv[++i]);
			cout << "Z Init Value Resetup to " << ZINITVAL << endl;
		}else if (!strcmp(argv[i], "-r")){
			RANGE = atoi(argv[++i]);
			cout << "Range Value Resetup to " << RANGE << endl;
		}
	}

	// Car information
	car_number = "";
	ifstream fin(CARINFOPATH);
	fin >> car_number;
	fin >> xml_log_path;
	fin >> fps_log_path;
	fin.close();

	// Log Print
	if (camLogLevel + sensorLogLevel > 0){
		cout << "Car Number : " << car_number << endl;
		cout << "XML Log Path : " << xml_log_path << endl;
		cout << "FPS Log Path : " << fps_log_path << endl;
	}

	// Date Sync..
	system("rdate -s time.bora.net 2> /dev/null");
	now_time = time(NULL);

	// Open I2C_ADC
	if (( i2cDev = open( i2cDevName, O_RDWR )) < 0 )
	{
		LogError( "Error  opening '%s': %s\n", i2cDevName, strerror( errno ));
		exit( 1 );
	}

	// Initialize kalma filter value
	for (int i = 0 ; i < NSENSOR ; i++ ) {
		P[i] = 1;
		Q = 1.0/100000.0;
		R = 0.0001;
		x[i] = 0.0;
	}

	// initialize I2C_ADC
	I2cSetSlaveAddress( i2cDev, 0x0B, I2C_USE_CRC );

	// Video Initialize
	videoBuffer = (struct vdIn *) calloc(1, sizeof(struct vdIn));
	if (init_videoIn(videoBuffer, (char *)DEV, WIDTH, HEIGHT, VIDEOFPS + FPS_OFFSET, V4L2_PIX_FMT_MJPEG, 1) < 0){
		cout << "Init video failure" << endl;
		exit(1);
	}

	// Buffer allocation
	videoBuffer->tmpbuffer = (unsigned char *) calloc(1, (size_t)videoBuffer->framesizeIn);
	g_buf = (unsigned char *) calloc(1, (size_t)videoBuffer->framesizeIn);

	// Run Thread (Capture start and Per Second Process Start)
	pthread_create(&cam_t, 0, cam_thread, NULL);
	pthread_detach(cam_t);
	pthread_create(&sec_t, 0, sec_thread, NULL);
	pthread_detach(sec_t);
	pthread_create(&fps_t, 0, fps_thread, NULL);
	pthread_detach(fps_t);
	pthread_create(&gps_t, 0, gps_thread, NULL);
	pthread_detach(gps_t);


	// Fps Handler Start
	act.sa_handler = div_sec;
	sigfillset(&(act.sa_mask));
	sigaction(SIGALRM, &act, NULL);
	ualarm(DIVSEC, DIVSEC);

	// unlimited loop for daemon
	while(!stop){
		if (getchar() == 'C')
			acciFlag = true;
	}

	ualarm(0,0);
	free(g_buf);
	free(videoBuffer);
	// Thread Destroy
	pthread_cond_destroy(&db_update);
	pthread_mutex_destroy(&db);
	pthread_cond_destroy(&fps_update);
	pthread_mutex_destroy(&fps);

	stop = true;
}
