#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>
#include <deque>
#include <unistd.h>
#include <pthread.h>
#include <fcntl.h>
#include <cerrno>
#include <csignal>

#include "v4l2uvc.h"
#include "utils.h"
#include "gps_parser.h"

#define I2CDEV "/dev/i2c-0"	//I2C DEV
#define NSENSOR 3			//The number of Sensor
#define FPS 12				//Sensoring FPS
#define VOLTPG		0.6		//Volt per Gravity
#define G			9.8		//Gravity Acceleration
// Default Value
#define DEFXINIT	315.5	//X axis Init(Standard) Value
#define DEFYINIT	360		//Y axis Init(Standard) Value
#define DEFZINIT	341.5	//Z axis Init(Standard) Value
#define DEFRANGE	20		//Range (Condition for accident)
#define VALPERVOLT 	(1024.0/5.0)	//Value per Volt
#define MPSFOMULA	G / VALPERVOLT / VOLTPG	// to Accelaration Fomula

#define DEV "/dev/video0"	//CAM DEV
#define VIDEOFPS 4			//CAM FPS
#define WIDTH 320			//Capture image width
#define HEIGHT 240			//Capture image height
#define FPS_OFFSET 3		//Optional value for more frequent capture

#define SAVETIME 60			//Log Time
#define USEC 1000000		//micro second
#define DIVSEC ((int)(USEC/FPS)) // 1 / FPS seconds

#define LOGPATH "./"		//Log file Path
#define CARINFOPATH "./CARINFO"	//Car Information File
#define XMLLOGEXT "xml"		//XML log file extension
#define CAMLOGEXT "tjpg" 	//Integrated capture log file extension
#define FPSLOGEXT "jpg"		//Capture Image extension
#define CARNOLENGTH 18		//Car number Length

// 1 Sec Data Structure
struct ONESEC{
	unsigned char *video;		//Capture Image
	size_t video_length;		//Capture Image length
	time_t time;				//Valid GPS time or System time
	double frame[FPS][NSENSOR]; //Sensoring information
	double speed;				//Speed
	struct llpoint gps_data;	//GPS Data
	int accident_flag;			//Accident Flag
};

// FPS Capture Image
struct CAMFPS{
	unsigned char *video;
	size_t video_length;
};

void *cam_thread(void *);
void *sec_thread(void *);
void *fps_thread(void *);
void div_sec(int);
void get_adc();
void make_capture_log(std::string);
void make_XML_log(std::string);
void make_fps_log(std::string);
bool is_accident_patten(double []);
void make_log();
