#include <sys/time.h> 
#include <sys/socket.h> 
#include <sys/signal.h> 
#include <sys/types.h> 
#include <sys/stat.h> 
#include <unistd.h> 
#include <stdlib.h> 
#include <stdio.h> 
#include <string.h> 
#include <fcntl.h>
#include <netinet/in.h> 
#include <arpa/inet.h> 
#include <termios.h>
#define GPSDEVICE "/dev/ttyS2"
#define MAXLINE 1
#define BAUDRATE B9600
#define PORTNUM 4400

//#define FD_SETSIZE    1024
int main(int argc, char **argv)
{
	int server_sockfd, client_sockfd, sockfd;

	//Serial port device
	int ttyfd;
	struct termios oldtio, newtio;

	// select 를 통해서 넘어오는 fd 의 갯수
	int fd_num;    

	int state, client_len;
	int i, maxi, maxfd;

	// 클라이언트로부터의 입력문을 저장
	char buf[MAXLINE];
	char line[MAXLINE];
	// client 연결을 저장하기 위한 배열
	int client[FD_SETSIZE];

	struct sockaddr_in clientaddr, serveraddr;
	fd_set readfds, allfds;
	
	pid_t	pid;

	if (( pid = fork()) < 0 )
		exit(0);

	else if ( pid != 0 )
		exit(0);
	
	chdir("/");
	
	fd_num = 0;

	//GPS device를 연다.
	//ttyfd = open_input_source("/dev/ttyS2");
	ttyfd = open(GPSDEVICE, O_RDWR | O_NOCTTY );
	if ( ttyfd < 0 )
	{
		perror(GPSDEVICE);
		exit(0);
	}

	tcgetattr(ttyfd,&oldtio); /* save current serial port settings */
	bzero(&newtio, sizeof(newtio)); /* clear struct for new port settings */

	/*
	BAUDRATE: 전송 속도. cfsetispeed() 및 cfsetospeed() 함수로도 세팅 가능
	CRTSCTS : 하드웨어 흐름 제어. (시리얼 케이블이 모든 핀에 연결되어 있는
	경우만 사용하도록 한다. Serial-HOWTO의 7장을 참조할 것.)
	CS8     : 8N1 (8bit, no parity, 1 stopbit)
	CLOCAL  : Local connection. 모뎀 제어를 하지 않는다.
	CREAD   : 문자 수신을 가능하게 한다.
	*/
	newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;

	/*
	IGNPAR   : Parity 에러가 있는 문자 바이트를 무시한다.
	ICRNL    : CR 문자를 NL 문자로 변환 처리한다. (이 설정을 안하면 다른
	컴퓨터는 CR 문자를 한 줄의 종료문자로 인식하지 않을 수 있다.)
	otherwise make device raw (no other input processing)
	*/
	newtio.c_iflag = IGNPAR | ICRNL;

	/*
	   Raw output.
	*/
	newtio.c_oflag = 0;
	/*
	ICANON   : canonical 입력을 가능하게 한다.
	disable all echo functionality, and don't send signals to calling program
	*/
	//newtio.c_lflag = ICANON;

	/*
	모든 제어 문자들을 초기화한다.
	디폴트 값은 <termios.h> 헤더 파일에서 찾을 수 있다. 여기 comment에도
	추가로 달아놓았다.
	*/

	//	newtio.c_cc[VINTR]    = 0;     /* Ctrl-c */
	//	newtio.c_cc[VQUIT]    = 0;     /* Ctrl-\ */
	//	newtio.c_cc[VERASE]   = 0;     /* del */
	//	newtio.c_cc[VKILL]    = 0;     /* @ */
	//	newtio.c_cc[VEOF]     = 4;     /* Ctrl-d */
	//	newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused */
	//	newtio.c_cc[VMIN]     = 1;     /* blocking read until 1 character arrives */
	//	newtio.c_cc[VSWTC]    = 0;     /* '\0' */
	//	newtio.c_cc[VSTART]   = 0;     /* Ctrl-q */
	//	newtio.c_cc[VSTOP]    = 0;     /* Ctrl-s */
	//	newtio.c_cc[VSUSP]    = 0;     /* Ctrl-z */
	//	newtio.c_cc[VEOL]     = 0;     /* '\0' */
	//	newtio.c_cc[VREPRINT] = 0;     /* Ctrl-r */
	//	newtio.c_cc[VDISCARD] = 0;     /* Ctrl-u */
	//	newtio.c_cc[VWERASE]  = 0;     /* Ctrl-w */
	//	newtio.c_cc[VLNEXT]   = 0;     /* Ctrl-v */
	//	newtio.c_cc[VEOL2]    = 0;     /* '\0' */
	
	/*
	   이제 modem 라인을 초기화하고 포트 세팅을 마친다.
	*/
	tcflush(ttyfd, TCIFLUSH);
	tcsetattr(ttyfd,TCSANOW,&newtio);


	// INET 소켓연결을 작성한다. 
	if ((server_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		perror("socket error : ");
		exit(0);
	}

	bzero(&serveraddr, sizeof(serveraddr));
	serveraddr.sin_family = AF_INET;
	serveraddr.sin_addr.s_addr = htonl(INADDR_ANY);
	serveraddr.sin_port = htons(PORTNUM);

	state = bind (server_sockfd, (struct sockaddr *)&serveraddr, 
			sizeof(serveraddr));

	if (state == -1)
	{
		perror("bind error : ");
		exit(0);
	} 

	state = listen(server_sockfd, 5);

	if (state == -1)
	{
		perror("listen error : ");                
		exit(0);
	}
	// INET 소켓작성 끝

	client_sockfd = server_sockfd;

	maxi = -1;

	// 현재 fd 의 최대값을 maxfd 에 입력한다. 
	maxfd = server_sockfd;

	// fd 배열을 -1 로 초기화 한다. 
	// 후에 accept 를 성공하면 해당 fd 로 배열을 채운다. 
	for (i = 0; i < FD_SETSIZE; i++)
	{
		client[i] = -1;
	}

	// INET 소켓 fd 에 대하여, fd_set 에 등록한다.
	FD_ZERO(&readfds);
	FD_SET(server_sockfd, &readfds);
	FD_SET(ttyfd, &readfds);

	while(1)
	{
		allfds = readfds;

		fd_num = select(maxfd + 1 , &allfds, (fd_set *)0, 
				(fd_set *)0, NULL);

		// 만약 소켓 파일지사자에 연결(읽기)이 들어온다면
		// accept 한다. 
		if (FD_ISSET(server_sockfd, &allfds))
		{
			client_len = sizeof(clientaddr);
			// accept 한다. 
			client_sockfd = accept(server_sockfd, 
					(struct sockaddr *)&clientaddr, &client_len);                

			// client 배열에 현재 연결된 client_sockfd 를 
			// 입력한다. 
			for (i = 0; i < FD_SETSIZE; i++)
			{
				if (client[i] < 0)
				{
					client[i] = client_sockfd;
					//printf("%d : %d\n", i, client_sockfd);
					break;
				}
			}

			// FD_SETSIZE 를 초과해서 client 가 접근할경우 
			// 연결을 종료한다. 
			if (i == FD_SETSIZE)
			{
				close(sockfd);
				client[i] = -1;
			}

			// fd_set 에 등록한다.
			FD_SET(client_sockfd,&readfds);

			if (client_sockfd > maxfd)
				maxfd = client_sockfd;

			if (i > maxi)
				maxi = i;

			if (--fd_num <= 0)
				continue;
		}

		else if ( FD_ISSET(ttyfd, &allfds)) {
			int n = read(ttyfd, line, MAXLINE);
			line[n] = 0;
			//printf("%s", line);
			
			for ( i = 0 ; i <= maxi ; i++ )
			{
				if ((sockfd = client[i]) < 0)
					continue;    

				write(sockfd, line, MAXLINE);
			}
		}

		for (i = 0; i <= maxi; i++)
		{
			if ((sockfd = client[i]) < 0)
				continue;    

			if (FD_ISSET(sockfd, &allfds))
			{
				if (read(sockfd, buf, MAXLINE) <= 0)
				{
					close(sockfd);
					FD_CLR(sockfd, &readfds);    
					client[i] = -1;
				}

				if (--fd_num <= 0)
					break;
			}
		}
	}
	tcsetattr(ttyfd,TCSANOW,&oldtio);
}
